#import "MooController.h"

// A misnomer... it's actually a path, as an NSString
#define MOO_EXEC_PATH_KEY			@"MOOExecutableURL"
#define MOO_LAST_DB_NAME_KEY		@"MOOLastDBName"
#define MOO_NEW_DB_NAME_FORMAT_KEY	@"MOONewDBNameFormat"
#define MOO_DATE_FORMAT_KEY			@"MOODateFormat"

#define MOO_DEFAULT_NEW_NAME_FORMAT	@"%@.new"
#define MOO_DEFAULT_DATE_FORMAT		@"%Y%m%d%H%M%S"

#define MOO_OLD_NAME_PLACEHOLDER		@"%@"
#define	MOO_DATE_PLACEHOLDER			@"%T"
#define	MOO_RANDOM_NUMBER_PLACEHOLDER	@"%R"

enum MooOpenPanelContextInfos
{
	MooPrefsContext,
	MooDbContext
};

@implementation MooController

+ (void)initialize
{
	// Set up the default name format
	if( [[NSUserDefaults standardUserDefaults] objectForKey:MOO_NEW_DB_NAME_FORMAT_KEY] == nil )
	{
		[[NSUserDefaults standardUserDefaults] setObject:MOO_DEFAULT_NEW_NAME_FORMAT
												  forKey:MOO_NEW_DB_NAME_FORMAT_KEY];
	}
	
	// Set up the default date format
	if( [[NSUserDefaults standardUserDefaults] objectForKey:MOO_DATE_FORMAT_KEY] == nil )
	{
		[[NSUserDefaults standardUserDefaults] setObject:MOO_DEFAULT_DATE_FORMAT
												  forKey:MOO_DATE_FORMAT_KEY];
	}
	
	srandomdev();
}

// init
- (id)init
{
    if (self = [super init])
    {
        [self setOutputString: [[NSAttributedString alloc] init]];
        [self setIsMooRunning: NO];
        [self setTaskStdError: nil];
        [self setTaskStdOut: nil];
        [self setTask: nil];
		
		shouldPromptForDb = NO;
    }
    return self;
}

//=========================================================== 
// dealloc
//=========================================================== 
- (void) dealloc
{
    [self setOutputString: nil];
    [self setTaskStdError: nil];
    [self setTaskStdOut: nil];
    [self setTask: nil];
    [self setDbPath: nil];
    [self setNewDbPath: nil];

    [super dealloc];
}


//=========================================================== 
//  outputString 
//=========================================================== 
- (NSAttributedString *) outputString { return [[outputString retain] autorelease]; }
- (void) setOutputString: (NSAttributedString *) anOutputString
{
    if (outputString != anOutputString)
    {
        [outputString release];
        outputString = [anOutputString retain];
    }
}

//=========================================================== 
//  isMooRunning 
//=========================================================== 
- (NSNumber*) isMooRunning { return [NSNumber numberWithBool:isMooRunning]; }
- (void) setIsMooRunning: (NSNumber*) flag
{
    isMooRunning = [flag boolValue];
}

//=========================================================== 
//  taskStdError 
//=========================================================== 
- (NSPipe *) taskStdError { return [[taskStdError retain] autorelease]; }
- (void) setTaskStdError: (NSPipe *) aTaskStdError
{
    if (taskStdError != aTaskStdError)
    {
        [taskStdError release];
        taskStdError = [aTaskStdError retain];
    }
}

//=========================================================== 
//  taskStdOut 
//=========================================================== 
- (NSPipe *) taskStdOut { return [[taskStdOut retain] autorelease]; }
- (void) setTaskStdOut: (NSPipe *) aTaskStdOut
{
    if (taskStdOut != aTaskStdOut)
    {
        [taskStdOut release];
        taskStdOut = [aTaskStdOut retain];
    }
}

//=========================================================== 
//  task 
//=========================================================== 
- (NSTask *) task { return [[task retain] autorelease]; }
- (void) setTask: (NSTask *) aTask
{
    if (task != aTask)
    {
        [task release];
        task = [aTask retain];
    }
}


//=========================================================== 
//  dbPath 
//=========================================================== 
- (NSString *) dbPath { return [[dbPath retain] autorelease]; }
- (void) setDbPath: (NSString *) aDbPath
{
    if (dbPath != aDbPath)
    {
        [dbPath release];
        dbPath = [aDbPath retain];
    }
}

//=========================================================== 
//  newDbPath 
//=========================================================== 
- (NSString *) newDbPath { return [[newDbPath retain] autorelease]; }
- (void) setNewDbPath: (NSString *) aNewDbPath
{
    if (newDbPath != aNewDbPath)
    {
        [newDbPath release];
        newDbPath = [aNewDbPath retain];
		
		// Update the user prefs with the last name
		[[prefsController defaults] setObject:newDbPath
									   forKey:MOO_LAST_DB_NAME_KEY];
    }
}

- (void)awakeFromNib
{
	// Start listening for data, even if we haven't created the pipes yet!
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(readDataNotification:)
												 name:NSFileHandleReadCompletionNotification
											   object:nil];

}

- (void)readDataNotification:(NSNotification*)aNotification
{
	if( [[aNotification name] compare:NSFileHandleReadCompletionNotification] == NSOrderedSame )
	{
		NSData *dataRead = [[aNotification userInfo] objectForKey:NSFileHandleNotificationDataItem];
		if( dataRead != nil )
		{
			NSString *newString = [NSString stringWithCString:[dataRead bytes] length:[dataRead length]];
			NSString *oldString = [[self outputString] string];
			[self setOutputString:
				[[NSAttributedString alloc] initWithString:[oldString stringByAppendingString:newString]]];
		}
		
		// Need to tell the file handle to keep reading
		id theObject = [aNotification object];
		if( [theObject isKindOfClass:[NSFileHandle class]] )
		{
			NSFileHandle *fileHandle = (NSFileHandle*)theObject;
			[fileHandle readInBackgroundAndNotify];
		}
	}
}

- (void)taskEndedNotification:(NSNotification*)aNotification
{
	if( ![task isRunning] )
	{
		[self setIsMooRunning:[NSNumber numberWithBool:NO]];
		[self setTask:nil];
	}
}

- (void)openPanelDidEnd:(NSOpenPanel *)panel returnCode:(int)returnCode  contextInfo:(void  *)contextInfo
{
	if( returnCode == NSOKButton )
	{
		switch((int)contextInfo)
		{
			case MooPrefsContext:
				// TODO: Should clean this up... hard to read
				[[prefsController defaults] setObject:[[panel filenames] objectAtIndex:0] forKey:MOO_EXEC_PATH_KEY];
				if( shouldPromptForDb )
				{
					[self promptUserToFindDb];
					shouldPromptForDb = NO;
				}
				break;
			
			case MooDbContext:
				[self setDbPath:[[panel filenames] objectAtIndex:0]];
				[self determineNewDbName];
				[self runMoo];
				break;
		}
	}
}

- (void)determineNewDbName
{
	// we're just going to have the name, not the full path
	NSMutableString *
	newStringFormat = [[[NSUserDefaults standardUserDefaults]
						objectForKey:MOO_NEW_DB_NAME_FORMAT_KEY]
							mutableCopy];

//#define MOO_OLD_NAME_PLACEHOLDER		@"%@"
//#define	MOO_DATE_PLACEHOLDER			@"%T"
//#define	MOO_RANDOM_NUMBER_PLACEHOLDER	@"%R"

	// Now replace the placeholders with the appropriate values
	NSRange dateRange = [newStringFormat rangeOfString:MOO_DATE_PLACEHOLDER];
	if( dateRange.location != NSNotFound )
	{
		NSString *dateFormat = [[NSUserDefaults standardUserDefaults] objectForKey:MOO_DATE_FORMAT_KEY];
		[newStringFormat replaceCharactersInRange:dateRange
									   withString:[[NSDate date] descriptionWithCalendarFormat:dateFormat
																					  timeZone:nil
																					    locale:nil]];
	}

	NSRange randomNumberRange = [newStringFormat rangeOfString:MOO_RANDOM_NUMBER_PLACEHOLDER];
	if( randomNumberRange.location != NSNotFound )
	{
		[newStringFormat replaceCharactersInRange:randomNumberRange
									   withString:[[NSNumber numberWithLong:random()] stringValue]];
	}

	NSRange oldNameRange = [newStringFormat rangeOfString:MOO_OLD_NAME_PLACEHOLDER];
	if( oldNameRange.location != NSNotFound )
	{
		NSString *oldName = [[self dbPath] lastPathComponent];
		[newStringFormat replaceCharactersInRange:oldNameRange withString:oldName];
	}

	[self setNewDbPath:newStringFormat];
}

- (void)runMoo
{
	// The newDbPath will be just a file name, in the same directory as dbPath
	NSArray *arguments = [NSArray arrayWithObjects:[self dbPath], [self newDbPath], nil];
	
	// Create pipes
	[self setTaskStdError:[NSPipe pipe]];
	[self setTaskStdOut:[NSPipe pipe]];
	
	// Create task
	[self setTask:[[NSTask alloc] init]];
	
	// Set up standard IO to our pipes
	[task setStandardError:[self taskStdError]];
	[[[self taskStdError] fileHandleForReading] readInBackgroundAndNotify];
	
	[task setStandardOutput:[self taskStdOut]];
	[[[self taskStdOut] fileHandleForReading] readInBackgroundAndNotify];
	
	// Keep an eye on the process
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(taskEndedNotification:)
												 name:NSTaskDidTerminateNotification
											   object:task];
	
	// Tell the task what executable to use
	[[self task] setLaunchPath:[[prefsController defaults] objectForKey:MOO_EXEC_PATH_KEY]];
	
	// Set the task's arguments
	[[self task] setArguments:arguments];
	
	// Set the working directory, so newDbPath makes sense
	[[self task] setCurrentDirectoryPath:[[self dbPath] stringByDeletingLastPathComponent]];
	
	// And away we go!
	[[self task] launch];
	
	[self setIsMooRunning:[NSNumber numberWithBool:YES]];
}

- (void)promptUserToFindDb
{
	NSOpenPanel *panel = [NSOpenPanel openPanel];
	
	[panel setAllowsMultipleSelection:NO];
	[panel setCanChooseFiles:YES];
	[panel setCanChooseDirectories:NO];
	[panel setMessage:@"Please choose a MOO database to use:"];
	
	NSString *lastDbName = [[prefsController defaults] objectForKey:MOO_LAST_DB_NAME_KEY];
	
	[panel beginSheetForDirectory:nil
							 file:lastDbName
							types:nil
				   modalForWindow:logWindow
				    modalDelegate:self
				   didEndSelector:@selector(openPanelDidEnd:returnCode:contextInfo:)
				      contextInfo:(void*)MooDbContext];
}

- (IBAction)chooseURL:(id)sender
{
	NSOpenPanel *panel = [NSOpenPanel openPanel];
	
	[panel setAllowsMultipleSelection:NO];
	[panel setCanChooseFiles:YES];
	[panel setCanChooseDirectories:NO];
	[panel setMessage:@"Please find the exectuble for the MOO:"];
	
	[panel beginSheetForDirectory:nil
							 file:nil
							types:nil
				   modalForWindow:prefsWindow
				    modalDelegate:self
				   didEndSelector:@selector(openPanelDidEnd:returnCode:contextInfo:)
				      contextInfo:(void*)MooPrefsContext];
}

- (IBAction)preferences:(id)sender
{
	[prefsWindow makeKeyAndOrderFront:self];
}

- (IBAction)startMoo:(id)sender
{
	// Get MOO exec path
	NSString *execPath = [[prefsController defaults] objectForKey:MOO_EXEC_PATH_KEY];
	
	if( execPath == nil || [execPath length] == 0 )
	{
		// What to do here? well, we should prompt the user
		shouldPromptForDb = YES;
		[self chooseURL:self];
	}
	else
	{
		[self promptUserToFindDb];
	}
}

- (IBAction)stopMoo:(id)sender
{
	if( task != nil )
	{
		[[self task] terminate];
		[[self task] waitUntilExit];
		[self setIsMooRunning:[NSNumber numberWithBool:NO]];
	}
}

@end
